program solcheck ( input,output,lnkone )
(* Checks that the guess produced by the LLL algorithm is actual solution. 
*)
; const
    maxd        = 500
  ; maxorbits   = 5000
  ; maxparams   = 5000
  ; maxrows     = 30000
  ; maxdim      = 500
  ; tab         = '     '

; type
    sign        = -1..1
  ; vector3     = array [1..3] of integer
  ; sgdvector3  = record x : vector3 ; sg : sign end
  ; pointer     = ^node
  ; node        = record par,val : integer ; next : pointer end
  ; colvector   = array [1..maxrows] of integer

; var
    lnkone :text
  ; d,orbits,params,rows,dim : integer
  ; gcdwithd,invmodd    : array [1..maxd] of integer
  ; orbit               : array [1..maxorbits] of
                          record
                            x : vector3
                          ; orbref,parref : integer
                          end
  ; rowptr              : array [1..maxrows] of pointer
  ; parofcol            : array [1..maxparams] of integer
  ; thecol              : colvector
  ; soln                : array [1..maxdim,1..maxparams] of integer
  ; freepar             : array [1..maxdim] of integer


(**********************************************************************)
(*                                                                    *)
(*                 basic routines involving integers                  *)
(*                                                                    *)
(**********************************************************************)

; function sgn ( a : integer ) : sign
  ; begin
      if a>0
      then sgn := 1
      else
      if a<0
      then sgn := -1
      else sgn := 0
    end (* sgn *)

; function modd ( a : integer ) : integer
  (* The smallest positive representative of a mod d. *)
  ; begin
      a := a mod d
    ; if a>0
      then modd := a
      else modd := a+d
    end (* modd *)

; procedure swapints ( var a,b : integer )
  (* Swap two integers. *)
  ; var
      h : integer
  ; begin
      h := a
    ; a := b
    ; b := h
    end (* swapints *)




(**********************************************************************)
(*                                                                    *)
(*                    making some auxiliary tables                    *)
(*                                                                    *)
(**********************************************************************)

; procedure make2tables
  (* Make a table containing the g.c.d. of every integer with d, *)
  (* and a table containing "inverse integers" of every integer. *)
  ; var
      a : integer
    ; b1,s1,b2,s2,q : integer
  ; begin
      for a := 1 to d do
      begin
        b1 := 0
      ; s1 := d (* b1*a cong. s1 mod d *)
      ; b2 := 1
      ; s2 := a (* b2*a cong. s2 mod d *)
      ; repeat
          q := s1 div s2
        ; b1 := b1-q*b2
        ; s1 := s1-q*s2
        ; swapints(b1,b2)
        ; swapints(s1,s2)
        until s2=0
      ; gcdwithd[a] := s1
      ; invmodd[a] := b1
      end
    ; for a := 1 to d do
      begin
        b1 := invmodd[a] mod (d div gcdwithd[a])
      ; if b1<=0 then b1 := b1+(d div gcdwithd[a])
      ; while gcdwithd[b1]>1 do b1 := b1+(d div gcdwithd[a])
      ; invmodd[a] := b1
      end
    end (* make2tables *)

(**********************************************************************)
(*                                                                    *)
(*           basic procedures involving (signed) 3-vectors            *)
(*                                                                    *)
(**********************************************************************)


; procedure makevecmodd ( x1,x2,x3 : integer ; var v : vector3 )
  (* Put x1,x2,x3 into a vector and reduce it modulo d. *)
  ; begin
      v[1] := modd(x1)
    ; v[2] := modd(x2)
    ; v[3] := modd(x3)
    end (* makevecmodd *)

; procedure plus ( v : vector3 ; var sv : sgdvector3 )
  (* Provide a 3-vector with a plus sign. *)
  ; begin
      sv.x := v
    ; sv.sg := 1
    end (* plus *)

; function prec ( v1,v2 : vector3 ) : boolean
  (* Test whether the first vector precedes the  *)
  (* second one in the lexicographical ordering. *)
  ; begin
      prec := false
    ; if v1[1]<v2[1]
      then prec := true
      else
        if v1[1]=v2[1] then
        if v1[2]<v2[2]
        then prec := true
        else
          if v1[2]=v2[2] then
          if v1[3]<v2[3]
          then prec := true
    end (* prec *)

; function oppo ( sv1,sv2 : sgdvector3 ) : boolean
  (* Test whether two signed 3-vectors are opposite. *)
  ; begin
      oppo := (sv1.x[1]=sv2.x[1]) and (sv1.x[2]=sv2.x[2])
        and (sv1.x[3]=sv2.x[3]) and (sv1.sg+sv2.sg=0)
    end (* oppo *)

; procedure multscalvec ( sc : integer ; sv : sgdvector3
  ; var sp : sgdvector3 )
  (* Multiply a signed 3-vector by a scalar. *)
  ; var
      i : integer
  ; begin
      for i := 1 to 3 do sp.x[i] := modd(sc*sv.x[i])
    ; sp.sg := sv.sg
    end (* multscalvec *)

; procedure movetox1 ( i : integer ; sv : sgdvector3
  ; var mv : sgdvector3 )
  (* Move x[i] to x[1]. *)
  ; begin
      if i>1 then
      with sv do
      begin
        swapints(x[i],x[1])
      ; sg := -sg
      end
    ; mv := sv
    end (* xitox1 *)
(**********************************************************************)
(*                                                                    *)
(*                 finding the minimal representative                 *)
(*                                                                    *)
(**********************************************************************)

; procedure actd8 ( sv : sgdvector3 ; var min : sgdvector3
                  ; var opp : boolean )
  (* Look at the orbit of sv under D8.  Return a   *)
  (* minimal element in min.  Return in opp        *)
  (* whether the orbit contains opposite elements. *)
  ; begin
      with sv do
      begin
        if modd(-x[2])<x[2] then x[2] := modd(-x[2])
      ; if modd(-x[3])<x[3] then x[3] := modd(-x[3])
      ; if x[3]<x[2] then
        begin
          swapints(x[3],x[2])
        ; sg := -sg
        end
      ; opp := ( x[2]=x[3] )
      end
    ; min := sv
    end (* actd8 *)

; procedure actd8xstab ( sv : sgdvector3 ; var min : sgdvector3
                         ; var opp : boolean )
  (* Look at the orbit of sv under                   *)
  (* D8 x ( stabilizer of x[1] in (Z/dZ)* ).         *)
  (* Return a minimal element in min.  Return in opp *)
  (* whether the orbit contains opposite elements.   *)
  ; var
      sc : integer
    ; sp,mini : sgdvector3
    ; oppi : boolean
  ; begin
      min := sv
    ; opp := false
    ; sc := 1 (* sc will run through integers cong. 1 mod d/sv.x[1] *)
    ; repeat
        if gcdwithd[sc]=1 then (* it's a unit *)
        begin
          multscalvec(sc,sv,sp)
        ; actd8(sp,mini,oppi)
        ; opp := opp or oppi or oppo(mini,min)
        ; if prec(mini.x,min.x) then min := mini
        end
      ; sc := sc+(d div sv.x[1])
      until sc>d
    end (* actd8xstab *)

; procedure acts4xmul ( sv : sgdvector3 ; var min : sgdvector3
                        ; var opp : boolean )
  (* Look at the orbit of sv under S4 x (Z/dZ)*.     *)
  (* Return a minimal element in min.  Return in opp *)
  (* whether the orbit contains opposite elements.   *)
  ; var
      i : integer
    ; sp,mv,mini : sgdvector3
    ; oppi : boolean
  ; begin
      with sv do
      begin
        min := sv
      ; opp := false
      ; for i := 1 to 3 do
        begin
          multscalvec(invmodd[x[i]],sv,sp)
        ; movetox1(i,sp,mv)
        ; actd8xstab(mv,mini,oppi)
        ; opp := opp or oppi or oppo(mini,min)
        ; if prec(mini.x,min.x) then min := mini
        end
      end
    end (* acts4xmul *)

(**********************************************************************)
(*                                                                    *)
(*                     using  the list of orbits                      *)
(*                                                                    *)
(**********************************************************************)

; function orbitnum ( v : vector3 ) : integer
  (* The number in the list of orbits of a minimal representative. *)
  ; var
      lw,md,up : integer
  ; begin
      lw := 1
    ; up := orbits+1
    ; repeat
        md := (lw+up) div 2
      ; if prec(v,orbit[md].x)
        then up := md
        else lw := md
      until up=lw+1
    ; orbitnum := lw
    end (* orbitnum *)

; procedure locate ( v : vector3 ; var on,pr : integer)
  (* Find the orbit number and parameter reference. *)
  ; var
      plusv,min : sgdvector3
    ; opp : boolean
  ; begin
      plus(v,plusv)
    ; acts4xmul(plusv,min,opp)
    ; on := orbitnum(min.x)
    ; pr := min.sg*orbit[on].parref
    end (* parrefofvec *)

(**********************************************************************)
(*                                                                    *)
(*                     making a list of variables                     *)
(*                                                                    *)
(**********************************************************************)

; procedure makeparams
  (* Make a list of parameters, i.e. of orbits whose function      *)
  (* values are not known to vanish or to be equal to the function *)
  (* value of some other orbit.  Set the parref field of orbit.    *)
  ; var
      i : integer
  ; begin
      params := 0
    ; for i := 1 to orbits do
      with orbit[i] do
      if abs(orbref)<i
      then
      begin
        if orbref<>0
        then parref := sgn(orbref)*orbit[abs(orbref)].parref
      end
      else
      begin
        params := params+1
      ; parref := params
      end
    end (* makeparams *)

(**********************************************************************)
(*                                                                    *)
(*                       input/output routines                        *)
(*                                                                    *)
(**********************************************************************)
 
; procedure readorbits
  ; var i : integer
  ; begin
    ; reset(lnkone,'lnkone') 
    ; readln(lnkone,d)
    ; readln(lnkone,orbits)
    ; for i := 1 to orbits do
      with orbit[i] do
      readln(lnkone,x[1],x[2],x[3],orbref)
    end

(**********************************************************************)
(*                                                                    *)
(*            elementary operations involving linked lists            *)
(*                                                                    *)
(**********************************************************************)

; procedure insert ( var head,prevptr,insptr,nextptr : pointer )
  (* Insert a new node into a linked list. *)
  ; begin
      new(insptr)
    ; if prevptr=nil
      then head := insptr
      else prevptr^.next := insptr
    ; insptr^.next := nextptr
    end (* insert *)

; procedure delete ( var head,prevptr,delptr : pointer )
  (* Delete a node from a linked list. *)
  ; var
      nextptr : pointer
  ; begin
      nextptr := delptr^.next
    ; dispose(delptr)
    ; if prevptr=nil
      then head := nextptr
      else prevptr^.next := nextptr
    end (* delete *)

; procedure advanceto ( var prevptr,ptr : pointer ; num : integer
                      ; var found : boolean )
  (* Advance the pointers ptr and prevptr until the par
     field of ptr is not smaller than num, or you reach
     the end of the list.  Return in found whether the
     par field of ptr eventually equals num *)
  ; var
      stop : boolean
  ; begin
      stop := false
    ; while (ptr<>nil) and not stop do
        if ptr^.par>=num
        then stop := true
        else
        begin
          prevptr := ptr
        ; ptr := ptr^.next
        end
    ; if ptr=nil
      then found := false
      else found := (ptr^.par=num)
    end (* advanceto *)

(**********************************************************************)
(*                                                                    *)
(*                    making the equations matrix                     *)
(*                                                                    *)
(**********************************************************************)

; procedure initeqnmatrix
  (* Initialize the equations matrix. *)
  ; var
      i : integer
  ; begin
      rows := 0
    ; for i := 1 to params do
      begin
        parofcol[i] := i
      end
    end (* initeqnmat *)

; procedure newrow
  ; begin
      rows := rows+1
    ; rowptr[rows] := nil
    end (* newrow *)

; procedure remlastrowifzero
  ; begin
      if rowptr[rows]=nil
      then rows := rows-1
    end (* remlastrowifzero *)

; procedure addpartolastrow ( num : integer )
  (* Add the parameter reference num to the
     last row of the equations matrix. *)
  ; var
      ptr,prevptr,newptr : pointer
    ; found : boolean
  ; begin
      if num<>0 then
      begin
        ptr := rowptr[rows]
      ; prevptr := nil
      ; advanceto(prevptr,ptr,abs(num),found)
      ; if found
        then
        begin
          ptr^.val := ptr^.val+sgn(num)
        ; if ptr^.val=0 then delete(rowptr[rows],prevptr,ptr)
        end
        else
        begin
          insert(rowptr[rows],prevptr,newptr,ptr)
        ; newptr^.par := abs(num)
        ; newptr^.val := sgn(num)
        end
      end
    end (* addpartolastrow *)

; procedure condeqn ( v1,v2,v3 : vector3 )
  (* Add the equation f(v1)+f(v2)+f(v3)=0 to the equations matrix,
     provided that the orbits of v2 and v3 vanish or do not precede
     that of v1. *)
  ; var
      on1,on2,on3,pr1,pr2,pr3 : integer
  ; begin
      locate(v1,on1,pr1)
    ; locate(v2,on2,pr2)
    ; locate(v3,on3,pr3)
    ; if ((pr2=0) or (on2>=on1)) and ((pr3=0) or (on3>=on1))
      then
      begin
        newrow
      ; addpartolastrow(pr1)
      ; addpartolastrow(pr2)
      ; addpartolastrow(pr3)
      ; remlastrowifzero
      end
    end (* condeqn *)

; procedure makeeqnmat
  (* Make a matrix with equations.  The equations are sorted by the
     lowest numbered orbit occurring in them which is not known to
     vanish. *)
  ; var
      i : integer
    ; v1,v2,v3 : vector3
  ; begin
      initeqnmatrix
    ; for i := 1 to orbits do
      with orbit[i] do
      if parref<>0 then (* otherwise the function vanishes *)
      begin
        makevecmodd(     x[1],     x[2],x[3],v1)
      ; makevecmodd(     x[2],x[1]-x[2],x[3],v2)
      ; makevecmodd(x[1]-x[2],     x[1],x[3],v3)
      ; condeqn(v1,v2,v3)
      ; makevecmodd(     x[1],     x[2],x[3],v1)
      ; makevecmodd(     x[2],x[1]+x[2],x[3],v2)
      ; makevecmodd(x[1]+x[2],     x[1],x[3],v3)
      ; condeqn(v1,v2,v3)
      ; makevecmodd(     x[2],     x[3],x[1],v1)
      ; makevecmodd(     x[3],x[2]-x[3],x[1],v2)
      ; makevecmodd(x[2]-x[3],     x[2],x[1],v3)
      ; condeqn(v1,v2,v3)
      ; makevecmodd(     x[2],     x[3],x[1],v1)
      ; makevecmodd(     x[3],x[2]+x[3],x[1],v2)
      ; makevecmodd(x[2]+x[3],     x[2],x[1],v3)
      ; condeqn(v1,v2,v3)
      ; makevecmodd(     x[3],     x[1],x[2],v1)
      ; makevecmodd(     x[1],x[3]-x[1],x[2],v2)
      ; makevecmodd(x[3]-x[1],     x[3],x[2],v3)
      ; condeqn(v1,v2,v3)
      ; makevecmodd(     x[3],     x[1],x[2],v1)
      ; makevecmodd(     x[1],x[3]+x[1],x[2],v2)
      ; makevecmodd(x[3]+x[1],     x[3],x[2],v3)
      ; condeqn(v1,v2,v3)
      end
    end (* makeeqnmat *)

(**********************************************************************)
(*                                                                    *)
(*                     Check that the lllresult is solution           *)
(*                                                                    *)
(**********************************************************************)

; function entry ( r,c : integer ) : integer
  (* Give entry (r,c) of the equations matrix. *)
  ; var
      ptr,prevptr : pointer
    ; found : boolean
  ; begin
      ptr := rowptr[r]
    ; prevptr := nil
    ; advanceto(prevptr,ptr,parofcol[c],found)
    ; if found
      then entry := ptr^.val
      else entry := 0
    end (* entry *)


; procedure checksolnspace
  ; var
      s,i,j,sum : integer
  ; begin
    ; for s := 1 to dim do
      begin
      ; for i := 1 to params do thecol[i]:=soln[s,i] 
      ; for i := rows downto 1 do
        begin
          sum := 0
        ; for j := 1 to params do
          if thecol[j]<>0 then sum := sum+entry(i,j)*thecol[j]
        ; if sum<>0 then 
          begin
          writeln('solution incorrect')
          ; soln[maxdim+abs(sum),maxdim+abs(sum)]:=0 (* cause 
                                                        runtime error*)
          end
        end
      end
   end (* calcsolnspace *)

(**********************************************************************)
(*                                                                    *)
(*                       input/output routines                        *)
(*                                                                    *)
(**********************************************************************)
 
; procedure readsubhint
  ; var
      i,s : integer
  ; begin
    ; readln(d)
    ; readln(params)
    ; readln(dim)
    ; for s := 1 to dim do
      begin
        readln(freepar[s])
      end
    ; for i := 1 to dim do
      for s := 1 to params do readln(soln[i,s])
    end

; procedure writesolnspace
  ; var
      i,s : integer
  ; begin
      writeln(d:1)
    ; writeln(params:1)
    ; writeln(dim:1)
    ; for s := 1 to dim do
      begin
        write(freepar[s]:1)
      ; if (s mod 10=0) or (s=dim)
        then writeln
        else write(tab)
      end
    ; for s := 1 to dim do
      for i := 1 to params do
      begin
        write(soln[s,i]:1)
      ; if (i mod 10=0) or (i=params)
        then writeln
        else write(tab)
      end
    end

; begin
    readorbits
  ; readsubhint
  ; make2tables
  ; makeparams
  ; makeeqnmat
  ; checksolnspace 
  ; writesolnspace
  end.

